from fpdf import FPDF
from opreport import objects, patching
import matplotlib.pyplot as plt
from matplotlib.ticker import MaxNLocator
import pandas as pd
import os

class PDF(FPDF):
    def __init__(self, client, start, end, bl_start, bl_end, logo, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.client = client
        self.start = start
        self.end = end
        self.bl_start = bl_start
        self.bl_end = bl_end
        self.logo = logo


    def header(self):
        logo_width = 25
        self.image(self.logo, 10, 8, h=13)
        self.set_font('Helvetica', 'B', 12)
        self.cell(w=self.epw, txt=self.client['name'], align='R', ln=1)
        self.ln(5)
        x = self.get_x()
        y = self.get_y()
        self.set_line_width(.5)
        self.line(x, y, self.epw+10, y)


        """
        self.set_font('Helvetica', 'B', 11)
        self.cell(40, 5, "Baseline:")   
        self.set_font('Helvetica', '', 11)
        self.cell(40, 5, self.bl_start.strftime("%m/%d/%Y") + " - " + self.bl_end.strftime("%m/%d/%Y") , ln=1)
        """

        self.ln(5)
 
        self.set_fill_color(254,231,205)
        self.set_font('Helvetica', 'B', 1)



    # Page footer
    def footer(self):
        if self.page_no() > 1:
            self.set_y(-15)
            self.set_font('Helvetica', '', 7)
            self.cell(0, 10, f'Page {self.page_no()}/{{nb}}', align='C')

def report(openv, tenants, daterange, outdir, logo):
    (start, end, bl_start, bl_end, utc_offset) = daterange
    for (tnum,tenant) in enumerate(tenants):
        #if tnum > 1:
        #    continue
        indexes = []
        data = []
        print(f'Start processing tenant {tenant["uniqueId"]} - {tenant["name"]}')
        resources = objects.get_objects(openv, obtype='resources', tenant=tenant['uniqueId'], queryString='agentInstalled:true+resourceType:Windows,Linux')
        resources = list(filter(lambda r: ('status' in r and r['status'] == 'UP'), resources))
        if len(resources) == 0:
            print("Skipping tenant as there are no Windows/Linux agent-installed resources.")
            continue
        resources_by_id = {}
        for resource in resources:
            resources_by_id[resource['id']] = resource
        all_patches = pd.DataFrame(pd.json_normalize(patching.get_patches_on_resources(openv, tenant['uniqueId'])))
        count_needs_updates = 0
        
        missing_patches = pd.DataFrame()
        if len(all_patches) > 0:
            all_patches['resource.state'] = all_patches.apply(lambda row: resources_by_id[row['resource.id']]['state'] if row['resource.id'] in resources_by_id else 'MISSING', axis=1)
            all_patches['resource.status'] = all_patches.apply(lambda row: resources_by_id[row['resource.id']]['status'] if row['resource.id'] in resources_by_id else 'MISSING', axis=1)
            missing_patches = all_patches[(all_patches['resource.status'] == 'UP') & (all_patches['patchStatus'] == 'MISSING')]
            count_needs_updates = missing_patches.groupby('resource.id')['resource.id'].nunique().count()

        count_current = len(resources) - count_needs_updates
        pct_compliance = "N/A"
        if len(resources) > 0:
            pct_complete = f'{int(count_current/len(resources)*100)}%'


        count = [count_current, count_needs_updates]
        labels = ['Current', 'Needs Updates']
        plt.bar(labels, count, color=['#4590f7','#ea3323'])
        
        ax=plt.gca()
        ax.set_ylabel("Device Count")
        ax.yaxis.set_major_locator(MaxNLocator(integer=True))
        ax.grid(axis='y')
        fig = ax.get_figure()
        fig.suptitle('Device Update Status', fontsize=24)
        plt.xticks(fontsize=16)
        plt.ylabel('Device Count', fontsize=16)
        imagefile = f'{tenant["uniqueId"]}_patch_device_update_status.png'
        plt.savefig(imagefile)
        plt.clf()
        plt.close('all')

        #print(json.dumps(opmetrics, indent=2, sort_keys=False))
        #sys.exit(0)
        pdf = PDF(tenant, start, end, bl_start, bl_end, logo, 'P', 'mm', 'Letter')

        # get total page numbers
        pdf.alias_nb_pages()

        #
        # COVER PAGE
        #
        pdf.add_page()
        pdf.set_font('Helvetica', 'B', 25)
        titlewidth = 170
        pdf.set_y(pdf.eph/3)

        pdf.cell(w=pdf.epw, txt='MONTHLY PATCH', align='C', ln=1)
        pdf.cell(w=pdf.epw, txt='MANAGEMENT REPORT', align='C', ln=1)
        pdf.ln(60)
        
        pdf.set_font('Helvetica', 'B', 16)
        pdf.cell(w=pdf.epw, txt=tenant['name'], align='C', ln=1)   
        pdf.ln(10)

        pdf.set_font('Helvetica', 'B', 14)
        pdf.cell(w=pdf.epw,txt="For the period:", align='C', ln=1)
        pdf.ln(5)
        pdf.cell(w=pdf.epw, txt=start.strftime("%m/%d/%Y") + " - " + end.strftime("%m/%d/%Y"), align='C', ln=1) 

        #
        # PIE CHART AND PCT
        #
        pdf.add_page()
        pdf.set_auto_page_break(auto=True, margin=15)
        pdf.set_font('Helvetica', '', 14)
        pdf.image(imagefile, w=pdf.epw/2, x=0)
        os.remove(imagefile)
        pdf.set_xy(pdf.epw/2, 26)
        pdf.cell(w=pdf.epw/2, txt="Devices That Are Up To Date", align='C', ln=1)
        pdf.ln(20)
        pdf.set_x(pdf.epw/2)
        pdf.set_font('Helvetica', '', 70)
        pdf.cell(w=pdf.epw/2, txt=pct_complete, align='C', ln=1)

        #
        # DEVICES MISSING REQUIRED PATCHES
        #

        cell_height = 6
        pdf.cell(w=pdf.epw, txt="", ln=1)
        pdf.ln(20)
        pdf.set_font('Helvetica', 'B', 13)
        pdf.cell(w=pdf.epw, txt="Devices Missing Required Patches", align='L', ln=1)
        pdf.ln(5)
        missing_per_server = missing_patches.groupby(['resource.name', 'resource.resourceType'])['resource.name'].count() if len(missing_patches)>0 else {}
        pdf.set_font('Helvetica', 'B', 11)
        col_headers = ['Device Name', 'System Type', 'Total Missing', 'Category']
        for (idx,col_header) in enumerate(col_headers):
            ln = 1 if idx == len(col_headers)-1 else 3
            align = 'C' if col_header=='Totel Missing' else 'L'
            pdf.multi_cell(w=pdf.epw/len(col_headers), h=cell_height, txt=col_header, align=align, border='B', ln=ln)  
        pdf.set_font('Helvetica', '', 11)     
        for (grp, val) in missing_per_server.items():           
            pdf.multi_cell(w=pdf.epw/len(col_headers), h=cell_height, txt=grp[0], align='L', border='TB', ln=3)
            pdf.multi_cell(w=pdf.epw/len(col_headers), h=cell_height, txt=grp[1], align='L', border='TB', ln=3)
            pdf.multi_cell(w=pdf.epw/len(col_headers), h=cell_height, txt=str(val), align='C', border='TB', ln=3)
            pdf.multi_cell(w=pdf.epw/len(col_headers), h=cell_height, txt="Security Updates", align='L', border='TB', ln=1)

        
        #
        # PATCH INSTALLATION SUMMARY
        #

        pdf.add_page()
        pdf.set_fill_color(38,68,121)
        pdf.set_text_color(255,255,255)
        pdf.set_font('Helvetica', 'B', 16)
        pdf.cell(w=pdf.epw, txt="Patch Installation Summary", align='L', ln=1, fill=True)
        pdf.ln(5)

        installed_patches = []
        if len(all_patches) > 0:
            all_patches['installedTime'] = pd.to_datetime(all_patches['installedTime'], infer_datetime_format=True)
            installed_patches = all_patches[(all_patches['patchStatus'] == 'INSTALLED') & (pd.notnull(all_patches['installedTime']))]
            installed_patches = installed_patches[(pd.to_datetime(installed_patches['installedTime']).dt.date>=start.date()) & (pd.to_datetime(installed_patches['installedTime']).dt.date<end.date())]
            print(f'Number of installed patches = {len(installed_patches)}')
            if len(installed_patches) > 0:
                sumlines = {}
                sumlines["Total Windows Patches Installed"] = len(installed_patches[(installed_patches['resource.resourceType']=='Windows')].groupby(['patch.id'])['patch.id'])
                sumlines["Total Linux Patches Installed"] = len(installed_patches[(installed_patches['resource.resourceType']=='Linux')].groupby(['patch.id'])['patch.id'])
                sumlines["Total Patches Installed"] = sumlines["Total Windows Patches Installed"] + sumlines["Total Linux Patches Installed"]

                sumlines["Total Windows Systems Patched"] = len(installed_patches[(installed_patches['resource.resourceType']=='Windows')].groupby(['resource.id'])['resource.id'])
                sumlines["Total Linux Systems Patched"]  = len(installed_patches[(installed_patches['resource.resourceType']=='Linux')].groupby(['resource.id'])['resource.id'])
                sumlines["Total Systems Patched"]  = sumlines["Total Windows Systems Patched"]  + sumlines["Total Linux Systems Patched"]

                sumlines["Total Patches Failed"] = 0
                sumlines["Total Patches Declined"] = 0

                patchtypes=installed_patches.groupby(['patch.category','patch.id']).size().groupby('patch.category').count()
                total = sum(patchtypes)
                patchtypes.plot.pie(labels=None, autopct=lambda s: str(round(s * total / 100)))
                ax=plt.gca()
                ax.set_ylabel(None)
                fig = ax.get_figure()
                fig.subplots_adjust(bottom=0.2)
                fig.suptitle('Patch Installation by Category', fontsize=17, fontname="Helvetica", fontweight="bold")
                plt.legend(patchtypes.index, loc="lower center", frameon=False, bbox_to_anchor=(0.5,-.2))
                imagefile = f'{tenant["uniqueId"]}_patch_types.png'
                plt.savefig(imagefile)
                plt.clf()
                plt.close('all')

                pdf.set_text_color(0,0,0)
                pdf.set_font('Helvetica', '', 12)
                pdf.cell(w=pdf.epw/2, txt="Patch Installation Summary", align='C', ln=1)
                pdf.ln(5)
                topy = pdf.get_y()
                pdf.set_font('Helvetica', '', 10)
                cell_height = 6
                for (name, val) in sumlines.items():
                    w_name = pdf.get_string_width(name)
                    w_val = pdf.epw/2-w_name-12
                    border = 'B'
                    if name == 'Total Patches Declined':
                        border = 'T'
                    pdf.cell(w=w_name, h=cell_height, txt=name, align='L', ln=0, border=border)
                    pdf.cell(w=w_val, h=cell_height, txt=str(val), align='R', ln=1, border=border)
                bottomy = pdf.get_y()
                pdf.image(imagefile, x=pdf.epw/2, y=topy-10, w=pdf.epw/2+10)
                os.remove(imagefile)
                pdf.ln(50)
                pdf.set_font('Helvetica', '', 12)
                pdf.cell(w=pdf.epw, txt="No Declined Patches During Report Period", align='C', ln=1)
                pdf.ln(20)
                pdf.set_font('Helvetica', '', 12)
                pdf.cell(w=pdf.epw, txt="No Failed Patch Installations During Report Period", align='C', ln=1)


        if len(installed_patches) == 0:
                pdf.set_text_color(0,0,0)
                pdf.set_font('Helvetica', '', 12)
                pdf.cell(w=pdf.epw, txt="No Patches Installed During Report Period", align='C', ln=1)            

        print(f'Done processing tenant {tenant["uniqueId"]} - {tenant["name"]}')

        repdir = f'{outdir}{os.sep}{tenant["name"]}'
        if not os.path.exists(repdir):
            os.makedirs(repdir)
        pdf_file = "patch_management_report.pdf"
        pdf.output(f'{repdir}{os.sep}{pdf_file}')


if __name__ == "__main__":
    # execute only if run as a script
    report()


import argparse
import sys
import opreport.helpers as helpers
from opreport import network_utilization_report, utilization_comparison_report, patch_management_report, remote_control_usage_report, hardware_inventory_report, software_inventory_report
from opreport import opsramp
import appdirs
import os
import traceback
import shutil

appauthor = "OpsRamp"
appname = "opreport"
default_conf_dir = appdirs.user_config_dir(appauthor, appname)
default_env_file = f'{default_conf_dir}{os.sep}environments.yml'
default_logo_file = f'{default_conf_dir}{os.sep}logo.jpeg'
default_output_dir = f'.{os.sep}output'
SAMPLE_ENVIRONMENTS = '''
#
# This file is used to provide named OpsRamp API credential sets
#


# Sample entries

- name: sample-environment-1
  url:  https://customername1.api.opsramp.com
  partner: msp_nnnnn                                    # If using a partner-level credential to do multi-client operations then use
  tenant: client_nnnnn                                  # the partner msp ID for both of these.
  client_id: oauth_client_id_from_integration
  client_secret: oauth_client_secret_from_integration
  vtoken: vtoken_from_integration                       # This line is only needed if sending alerts to an inbound webhook integration

- name: sample-environment-12
  url:  https://customername2.api.opsramp.com
  partner: msp_nnnnn
  tenant: client_nnnnn
  client_id: oauth_client_id_from_integration
  client_secret: oauth_client_secret_from_integration
  vtoken: vtoken_from_integration

#
# Add your entries below here
#
'''

def do_arg_parsing():
    parser = argparse.ArgumentParser(description='Multi-Client Reporting Utility for OpsRamp')
    subparsers = parser.add_subparsers(required=True, help='Available commands', dest='command')

    parser_uc = subparsers.add_parser('uc', help='Generate Server Utilization Comparison Reports')
    parser_uc.add_argument('--baseline_months_back', type=int, default=1, help='How many months back to use for the baseline data')
    subparsers.add_parser('nu', help='Generate Network Utilization  Reports')
    subparsers.add_parser('pm', help='Generate Patch Management Reports')
    subparsers.add_parser('rcu', help='Generate Remote Control Usage Reports')

    parser_hi = subparsers.add_parser('hi', help='Generate Hardware Inventory Reports')
    parser_hi.add_argument('--includefields', help='Comma-delimited list of fields to include in the report')
    parser_hi.add_argument('--query', default='type:DEVICE', help='Resource query string as per https://develop.opsramp.com/resource-management/tenants-tenantid-resources-search (Default: "type:DEVICE")')
    parser_hi.add_argument('--listfields', action='store_true', help='Show a list of available fields.  Report will not be generated.')
    parser_hi.add_argument('--extrafields', action='store_true', help='Include additional detailed fields.  This will run slower as a get for each individual resource must be done.')
    parser_hi.add_argument('--format', help='Output format', choices = ['excel', 'pdf'], default = 'excel')

    parser_si = subparsers.add_parser('si', help='Generate Software Inventory Reports')
    parser_si.add_argument('--query', default='agentInstalled:true', help='Resource query string as per https://develop.opsramp.com/resource-management/tenants-tenantid-resources-search (Default: "type:DEVICE")')

    envfile_creds = parser.add_argument_group(title='Use yaml environment file credentials')
    cli_creds = parser.add_argument_group(title='Use command line credentials')

    envfile_creds.add_argument('--env', help='Name of environment to use, referencing a named set of API credentials in environments.yml')
    envfile_creds.add_argument('--envfile', default=default_env_file, help=f'Location of environments YAML file to be used (Default: {appdirs.user_config_dir(appauthor, appname)}{os.sep}environments.yml)')

    cli_creds.add_argument('--api_url', help='Customer-specific API URL such as https://mycompany.api.opsramp.com')
    cli_creds.add_argument('--key', help='Authentication key')
    cli_creds.add_argument('--secret', help='Authentication secret')
    cli_creds.add_argument('--tenant', help='Authentication MSP or Tenant ID')
    

    parser.add_argument('--start', help='Start date of reporting period in YYYY-MM-DD format (default=1st of prior month)')
    parser.add_argument('--end', help='End date of reporting period in YYYY-MM-DD format (default=Last day of prior month)')
    parser.add_argument('--outdir', default=default_output_dir, help=f'Directory where report output will be written (Default: {default_output_dir})')
    parser.add_argument('--logo', default=default_logo_file, help=f'Logo image file for report headers (Default: {default_logo_file})')
    parser.add_argument('--secure', default=True, help='Whether or not to verify SSL cert (Default: True)')
    parser.add_argument('--tenant_criteria', help='Search criteria for which tenants to include as per https://develop.opsramp.com/tenancy-access-controls/tenants-orgid-clients-search (Default: "activeStatus:true", which will include all active tenants)')

    args = parser.parse_args()
    return args



def main():
    
    try:
        if not (len(sys.argv) == 2 and sys.argv[1] == '-h'):
            try:
                f = open(default_env_file, 'r')
                f.close()
            except FileNotFoundError:
                print(f'\nCould not find default environments file {default_env_file}, so creating it with sample content.\nPlease update it with your OpsRamp partner-level API credentials.')
                print('Run again with -h option to see available options and commands.\n')
                os.makedirs(default_conf_dir)
                env_file = open(default_env_file, "w")
                env_file.write(SAMPLE_ENVIRONMENTS)
                env_file.close()
                shutil.copy(f'{os.path.dirname(__file__)}{os.sep}..{os.sep}conf{os.sep}sample-logo.jpeg', default_logo_file)
                sys.exit(0)

        args = do_arg_parsing()

        if args.env is None and (args.api_url is None or args.key is None or args.secret is None or args.tenant is None) :
            print('For authentication, either the --env option referencing to a named set of API credentials in environments.yml, or all of: --api_url, --key, --secret, --tenant must be provided.')
            sys.exit(1)

        if args.env is not None and (args.api_url is not None or args.key is not None or args.secret is not None or args.tenant is not None) :
            print('For authentication, use either the --env option referencing to a named set of API credentials in environments.yml, or --api_url, --key, --secret, and --tenant - not both.')
            sys.exit(1)
        
        if args.env is None:
            print('\nUsing API credentials provided in command line')

        else:
            print(f'\nGetting environment named {args.env} from {args.envfile}')

        print(f'Report output will be sent to {args.outdir}')

        openv = opsramp.OpsRampEnv(args.env, args.api_url, args.key, args.secret, None, args.tenant, args.envfile, args.secure)
        tenants = openv.get_tenants(queryString=args.tenant_criteria)
        dateranges = helpers.compute_dates(args)

        
        if args.command == 'uc':
            utilization_comparison_report.report(openv, tenants, dateranges, args.outdir, args.logo)
        elif args.command == 'nu':
            network_utilization_report.report(openv, tenants, dateranges, args.outdir, args.logo)
        elif args.command == 'pm':
            patch_management_report.report(openv, tenants, dateranges, args.outdir, args.logo)
        elif args.command == 'rcu':
            remote_control_usage_report.report(openv, tenants, dateranges, args.outdir, args.logo)
        elif args.command == 'hi':
            includefields = None
            if args.listfields:
                fieldlist = hardware_inventory_report.getfields(openv, tenants, args.extrafields)
                print('\nThe following fields can be specified:')
                [print(f'  {field}') for field in fieldlist]
            else:
                if args.includefields is not None:
                    includefields = [fldname.strip() for fldname in args.includefields.split(',')]
                hardware_inventory_report.report(openv, tenants, includefields, args.extrafields, args.format, args.query, args.outdir, args.logo)
        elif args.command == 'si':
            software_inventory_report.report(openv, tenants, args.query, args.outdir, args.logo)
        print("Done")

    except Exception as err:
        print('\nAn error has occurred:\n')
        print(f"{err.args=}\n")
        print("-"*60)
        traceback.print_exc(file=sys.stdout)
        print("-"*60)




if __name__ == "__main__":
    # execute only if run as a script
    main()



def get_objects(openv, obtype, queryString=None, searchQuery=None,countonly=False, attrId=None, tenant=None, partner=None):
    if tenant is None:
        tenant = openv.env['tenant']
    if partner is None:
        partner = openv.env['partner']

    endpoints = {
        "clients": f'/api/v2/tenants/{partner}/clients/tenant',
        "incidentCustomFields": f'/api/v2/tenants/{tenant}/customFields/INCIDENT',
        "deviceGroups": f'/api/v2/tenants/{tenant}/deviceGroups/minimal',
        "users": f'/api/v2/tenants/{tenant}/users/search',
        "userGroups": f'/api/v2/tenants/{tenant}/userGroups',
        "urgencies": f'/api/v2/tenants/{tenant}/incidents/urgencies',
        "customAttributes": f'/api/v2/tenants/{tenant}/customAttributes/search',
        "resources": f'/api/v2/tenants/{tenant}/resources/search',
        "resourcesNewSearch": f'/api/v2/tenants/{tenant}/query/execute',
        "assignedAttributeEntities": f'/api/v2/tenants/{tenant}/customAttributes/{str(attrId)}/assignedEntities/search',
        "consoleRecordings": f'/api/v2/tenants/{tenant}/resources/auditRecordings/search'
    }

    url = endpoints[obtype]

    return openv.do_get(url, 1, queryString, searchQuery, countonly, attrId)
    
def get_object(openv, obtype, id, tenant=None, partner=None):
    if tenant is None:
        tenant = openv.env['tenant']
    if partner is None:
        partner = openv.env['partner']

    endpoints = {
        "resource": f'/api/v2/tenants/{tenant}/resources/{id}',
        "discoveredApps": f'/api/v2/tenants/{tenant}/resources/{id}/applications',
        "discoveredServices": f'/api/v2/tenants/{tenant}/resources/{id}/discoveredServices',
    }

    url = endpoints[obtype]

    paginate = True
    pagebug = False
    if obtype == 'discoveredApps':
        paginate = False
    if obtype == 'discoveredServices':
        pagebug = True
    return openv.do_get(url, paginate=paginate, pagebug=pagebug)


import requests

def get_metricdatasets(openv, tenant, resource, metric, startTime, endTime):
    if tenant is None:
        tenant = openv.env['tenant']
    url = f'/api/v2/metric/search?tenant={tenant}&resource={resource}&timeseries_type=RealTime&metric={metric}&startTime={str(startTime)}&endTime={str(endTime)}'
    return openv.do_get(url)

def query_metrics(openv, tenant, resource, metric, startTime, endTime):
    url = f'/metricsql/api/v3/tenants/{tenant}/metrics&resource={resource}&timeseries_type=RealTime&metric={metric}&startTime={str(startTime)}&endTime={str(endTime)}'
    return openv.do_get(url)

def promql_instance_search(openv, tenant=None, query=None):
    if tenant is None:
        tenant = openv.env['tenant']
    headers = {
        'Content-Type'      : 'application/json',
        'Accept'            : 'application/json',
        'Authorization'     : 'Bearer ' + openv.get_token()
    }

    url = f'{openv.env["url"]}/metricsql/api/v7/tenants/{tenant}/metrics/latest'
    data = {
        "query": query
    }

    response = requests.request("POST", url, headers=headers, verify=openv.isSecure, json=data)
    
    try:
        responseobj = response.json()
    except Exception as e:
        print(repr(response))
        return None

    if 'status' in responseobj and responseobj['status'] == 'success':
        return responseobj['data']['result']

    return None


import opreport.helpers as helpers
import opreport.objects as objects
from datetime import datetime
import pandas as pd
import matplotlib.pyplot as plt
import  matplotlib.dates as mdates
import os
import math
import re
import xlsxwriter

includefields = [1,2,3,4,5,6,7]

def report(openv, tenants, queryString, outdir, logo):
    for tenant in tenants:
        print("Tenant is " + tenant['uniqueId'] + " - " + tenant['name'])

        #
        # Initialize pdf doc and header/footer
        #
        pw = max(len(includefields)*40, 216) # min of 8.5 in, but expands based on # columns
        ph = 279 # 11 in.
        pdf = helpers.PDF_STD_HEADER_FOOTER(tenant, "Software Inventory Report", None, None, None, None, logo, 'P', 'mm', (pw,ph))
        pdf.alias_nb_pages()
        pdf.set_auto_page_break(auto=True, margin=15)
        pdf.add_page()

        #
        # Get resources
        #

        if queryString is None:
            queryString = 'agentInstalled:true'
        resources = objects.get_objects(openv, 'resources', tenant=tenant['uniqueId'], queryString=queryString)

        for ridx,resource in enumerate(resources):
            apps = objects.get_object(openv, 'discoveredApps', resource['id'], tenant['uniqueId'])
            services = objects.get_object(openv, 'discoveredServices', resource['id'], tenant['uniqueId'])
            df_apps = pd.DataFrame(pd.json_normalize(apps))
            df_services = pd.DataFrame(pd.json_normalize(services))


            #
            # Add the apps pdf table
            #
            if ridx > 0:
                pdf.add_page()
            pdf.set_font('Helvetica', 'B', 11)
            pdf.cell(w=pdf.epw, txt=f'Installed applications for {resource["name"]}', align='L', ln=1)
            pdf.ln(5)
            if len(df_apps) > 0:
                for fld in ['installedDate', 'createdDate', 'modifiedDate']:
                    df_apps[fld] = pd.to_datetime(df_apps[fld]).dt.strftime('%Y-%m-%d')
                helpers.pdf_table_from_df(pdf, df_apps, df_apps.columns)
            else:
                pdf.ln(10)
                pdf.cell(w=pdf.epw, txt="No installed applications discovered for this device.", align='C', ln=1)

            #
            # Add the services pdf table
            #
            pdf.add_page()
            pdf.set_font('Helvetica', 'B', 11)
            pdf.cell(w=pdf.epw, txt=f'Installed services for {resource["name"]}', align='L', ln=1)
            pdf.ln(5)
            if len(df_services) > 0:
                #for fld in ['installedDate', 'createdDate', 'modifiedDate']:
                #    df_services[fld] = pd.to_datetime(df_services[fld]).dt.strftime('%Y-%m-%d')
                helpers.pdf_table_from_df(pdf, df_services, df_services.columns)
            else:
                pdf.ln(10)
                pdf.cell(w=pdf.epw, txt="No installed services discovered for this device.", align='C', ln=1)


        #
        # Prep the output directory
        #
        repdir = f'{outdir}{os.sep}{tenant["name"]}'
        if not os.path.exists(repdir):
            os.makedirs(repdir)

        #
        # Write the pdf file
        #
        pdf_file = "software_inventory_report.pdf"
        pdf.output(f'{repdir}{os.sep}{pdf_file}')


if __name__ == "__main__":
    # execute only if run as a script
    report() 


import opreport.helpers as helpers
import opreport.objects as objects
from datetime import datetime
import pandas as pd
import matplotlib.pyplot as plt
import  matplotlib.dates as mdates
import os
import math
import re
import xlsxwriter

ALL_COLUMNS = {
    "id": "Resource ID",

}

DEFAULT_FIELD_LIST = [
    'client.name',
    'name',
    'resourceType',
    'make',
    'model',
    'ipAddress',
    'osName',
    'serialNumber'
]

DEFAULT_HEADERS = {
    'client.name': 'Customer Name',
    'id': 'ID',
    'name': 'name',
    'resourceType': 'Resource Type',
    'make': 'Make',
    'model': 'Model',
    'ipAddress': 'IP Address',
    'osName': 'OS',
    'serialNumber': 'Serial Number'
}

def getfields(openv, tenants, extrafields):
    df = []
    for tenant in tenants:
        queryString = 'type:DEVICE'
        resources = objects.get_objects(openv, 'resources', tenant=tenant['uniqueId'], queryString=queryString)
        if len(resources)==0:
            continue
        if extrafields:
            resource_detail = objects.get_object(openv, 'resource', resources[0]['id'], tenant=tenant['uniqueId'])
            resources[0] = {**resources[0], **resource_detail}
        df = pd.DataFrame(pd.json_normalize(resources))
        for column in df.columns:
            try:
                df = df.drop([f'generalInfo.{column}'], axis=1)
            except:
                pass
        df.columns = df.columns.str.replace('generalInfo.', '', regex=False)
        break        
    retval =  df.columns.values
    retval.sort()
    return retval

def get_cpu_count(cpus):
    if cpus == 'nan':
        numcpus = ''
    else:
        numcpus = len(eval(cpus))
        if numcpus == 0:
            numcpus = ''
    return numcpus

def report(openv, tenants, includefields, extrafields, format, queryString, outdir, logo):
    for tenant in tenants:
        print("Tenant is " + tenant['uniqueId'] + " - " + tenant['name'])

        #
        # Get resources
        #

        if includefields is None:
            includefields = DEFAULT_FIELD_LIST
        if queryString is None:
            queryString = 'type:DEVICE'
        resources = objects.get_objects(openv, 'resources', tenant=tenant['uniqueId'], queryString=queryString)
        resources_extra = []
        if extrafields:
            for ridx, resource in enumerate(resources):
                print(f'  Getting extra field details for {resource["name"]}')
                resource_detail = objects.get_object(openv, 'resource', resource['id'], tenant=tenant['uniqueId'])
                resources[ridx] = {**resources[ridx], **resource_detail}
                resources_extra.append(resource_detail)


        #
        # INITIALIZE PDF DOC
        #

        if format == 'pdf' :
            pw = max(len(includefields)*40, 216) # min of 8.5 in, but expands based on # columns
            ph = 279 # 11 in.
            pdf = helpers.PDF_STD_HEADER_FOOTER(tenant, "Hardware Inventory Report", None, None, None, None, logo, 'P', 'mm', (pw,ph))
            pdf.alias_nb_pages()
            pdf.set_auto_page_break(auto=True, margin=15)
            pdf.add_page()



        df = pd.DataFrame(pd.json_normalize(resources))
        for column in df.columns:
            try:
                df = df.drop([f'generalInfo.{column}'], axis=1)
            except:
                pass
        df.columns = df.columns.str.replace('generalInfo.', '', regex=False)


        repdir = f'{outdir}{os.sep}{tenant["name"]}'
        if not os.path.exists(repdir):
            os.makedirs(repdir)
        if format == 'pdf' :
            if len(df) > 0:
                pdf.set_font('Helvetica', 'B', 11)
                pdf.cell(w=pdf.epw, txt=f'Hardware Inventory Details', align='L', ln=1)
                pdf.ln(5)

                pdf.set_fill_color(224, 102, 54)
                pdf.set_text_color(255,255,255)
                pdf.set_font('Helvetica', 'B', 9)
                line_height = pdf.font_size * 4
                col_width = pdf.epw / len(includefields) # distribute content evenly as baseline
                for colidx,column in enumerate(includefields):
                    border = 'TB'
                    ln = 3
                    if colidx == 0:
                        border = 'LTB'
                    elif colidx == len(includefields)-1:
                        border = 'TBR'
                        ln = 1
                    pdf.multi_cell(col_width, line_height, column, border=border, ln=ln, max_line_height=pdf.font_size, fill=True, align='C')

                pdf.set_text_color(0,0,0)
                pdf.set_font('Helvetica', '', 7)
                line_height = pdf.font_size * 2
                iter = df.iterrows()
                try:
                    while True:
                        nextrec = next(iter)
                        idx = nextrec[0]
                        rec = nextrec[1]
                        if idx%2 == 0:
                            pdf.set_fill_color(252, 245, 233)
                        else:
                            pdf.set_fill_color(249, 234, 213)

                        numrowsmax = int(math.ceil(max([pdf.get_string_width(str(rec[column]))/col_width for column in includefields if not pd.isna(rec[column])])))
                        this_line_height = line_height*numrowsmax
                        for colidx, column in enumerate(includefields):
                            border = 'TB'
                            ln = 3
                            if colidx == 0:
                                border = 'LTB'
                            elif colidx == len(includefields)-1:
                                border = 'TBR'
                                ln = 1
                            colval = rec[column]
                            if pd.isna(colval):
                                colval = ""
                            else:
                                colval = str(colval)
                            if column == 'osName':
                                colval = re.sub(r' [.\d]+ Build.*','', colval)
                            pdf.multi_cell(col_width, this_line_height, colval, border=border, ln=ln, max_line_height=pdf.font_size, fill=True, align='C')
                except StopIteration:
                    pass   
            else:
                pdf.ln(60)
                pdf.cell(w=pdf.epw, txt="No Devices found for this client.", align='C')

            pdf_file = "hardware_inventory_report.pdf"
            pdf.output(f'{repdir}{os.sep}{pdf_file}')

        elif format == 'excel':
            dfout = pd.DataFrame(columns=includefields)
            if len(df) > 0:
                for cidx,column in enumerate(includefields):
                    if column in df.columns:
                        if column == 'cpus':
                            cpudf = pd.json_normalize(pd.json_normalize(df['cpus'])[0])
                            if 'processorName' in cpudf.columns:
                                dfout['cpus'] = cpudf['processorName']
                            else:
                                dfout['cpus'] = ''
                            if 'numberOfCores' in cpudf.columns:
                                dfout.insert(cidx+1, 'cpus.cores', cpudf['numberOfCores'])
                            else:
                                dfout.insert(cidx+1, 'cpus.cores', '')
                            dfout.insert(cidx+1, 'cpus.count', df['cpus'].astype(str).apply(get_cpu_count))

                        else:
                            dfout[column] = df[column]
            filename = "hardware_inventory_report.xlsx"
            writer = pd.ExcelWriter(f'{repdir}{os.sep}{filename}', engine="xlsxwriter")
            dfout.to_excel(writer, index=False, sheet_name='Hardware Inventory')
            workbook  = writer.book
            worksheet = writer.sheets['Hardware Inventory']
            for i, col in enumerate(dfout.columns):
                column_len = max(min(max(dfout[col].astype(str).str.len().max(), len(col) + 2),100),20)
                if pd.isna(column_len):
                    column_len = 20
                worksheet.set_column(i, i, column_len)
            writer.save()




if __name__ == "__main__":
    # execute only if run as a script
    report() 


from fpdf import FPDF
import opreport.metrics as metrics
import matplotlib.pyplot as plt
import  matplotlib.dates as mdates
import pandas as pd
import os
import time
from datetime import datetime
from dateutil.relativedelta import relativedelta
import math
import re


class PDF_STD_HEADER_FOOTER(FPDF):
    def __init__(self, tenant, title, start, end, bl_start=None, bl_end=None, logo_file='logo.jpeg', *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.resource = {'name': 'none'}
        self.tenant = tenant
        self.start = start
        self.end = end
        self.bl_start = bl_start
        self.bl_end = bl_end
        self.title = title
        self.logo_file = logo_file


    def header(self):
        if self.page_no() == 1:
            self.image(self.logo_file, 10, 8, h=13)
            self.set_font('Helvetica', 'B', 16)
            self.ln(15)
            self.cell(30, 5, self.title, border=False, ln=1, align='L')
            self.ln(5)
            self.set_font('Helvetica', 'B', 11)
            self.cell(40, 5, "Customer:")
            self.set_font('Helvetica', '', 11)
            self.cell(40, 5, self.tenant['name'], ln=1)

            if self.start is not None and self.end is not None:
                self.set_font('Helvetica', 'B', 11)
                self.cell(40, 5, "Period:")
                self.set_font('Helvetica', '', 11)
                self.cell(40, 5, self.start.strftime("%m/%d/%Y") + " - " + self.end.strftime("%m/%d/%Y") , ln=1)
            else:
                self.set_font('Helvetica', 'B', 11)
                self.cell(40, 5, "Report Date:")
                self.set_font('Helvetica', '', 11)
                self.cell(40, 5, datetime.now().strftime("%m/%d/%Y"), ln=1)                

            if self.bl_start is not None and self.bl_end is not None:
                self.set_font('Helvetica', 'B', 11)
                self.cell(40, 5, "Baseline:")   
                self.set_font('Helvetica', '', 11)
                self.cell(40, 5, self.bl_start.strftime("%m/%d/%Y") + " - " + self.bl_end.strftime("%m/%d/%Y") , ln=1)

            self.ln(5)
        else:
            logo_width = 25
            self.image(self.logo_file, 10, 8, h=9)
            self.set_font('Helvetica', 'B', 12)
            self.cell(w=self.epw, txt=self.tenant['name'], align='R', ln=1)
            self.ln(5)
            x = self.get_x()
            y = self.get_y()
            self.set_line_width(.5)
            self.line(x, y, self.epw+10, y)
            self.ln(5)

    # Page footer
    def footer(self):
        self.set_y(-15)
        self.set_font('Helvetica', '', 7)
        self.cell(0, 10, f'Page {self.page_no()}/{{nb}}', align='C')

def top_n_timeseries_plus_detail_with_instance(openv, tenant, pdf, title, metric, range, span, offset=None, n=10, top_aggtype='avg', fig_h=3, fig_w=8, dpi=600, show_chart=True, show_details=True, all_details=True):
    if offset is None:
        offsetstring = ""
    else:
        offsetstring = f' offset {offset}'
    search = f'topk({n},{top_aggtype}_over_time({metric}[{range}:{span}]{offsetstring}))'
    top_n = metrics.promql_instance_search(openv, tenant['uniqueId'], search)
    labels = []
    labelmatch = ""
    for (idx,instance) in enumerate(top_n):
        if idx > 0:
            labelmatch += '|'
        labelmatch += instance['metric']['uuid']
        labels.append(instance['metric']['name'])
    
    if (show_chart):
        search = f'{metric}{{uuid=~"{labelmatch}"}}[{range}:{span}]{offsetstring}'
        data = metrics.promql_instance_search(openv, tenant['uniqueId'], search)
        df = []
        for (idx,topn) in enumerate(top_n):
            for (idx2, row) in enumerate(data):
                if topn['metric']['name'] == row['metric']['name'] and topn['metric']['instance'] == row['metric']['instance']:
                    device_name = row['metric']['name']
                    instance_name = row['metric']['instance']
                    colname = device_name + " (" + instance_name + ")"
                    dfpart = pd.DataFrame(row['values'], columns=['timestamp', colname])
                    dfpart[colname] = pd.to_numeric(dfpart[colname])
                    if idx == 0:
                        df = dfpart
                    else:
                        df = pd.merge(df, dfpart, on='timestamp', how='outer')
        if len(df) > 0:
            df=df.set_index('timestamp')
            #df = pd.merge(pd.DataFrame(index=(pd.date_range(pdf.start,pdf.end).astype(int)/1000000000).astype(int)[:-1]),df, how='outer', left_index=True, right_index=True)
            df = df.sort_index()
            plt.figure(figsize=(fig_w,fig_h), dpi=dpi)
            for col in df.columns.to_list():
                plt.plot_date(mdates.epoch2num(df.index), df[col], fmt='-')
            ax=plt.gca()
            fig = ax.get_figure()
            fig.subplots_adjust(bottom=0.28)
            fig.legend(df.columns, loc="lower center", ncol=4, frameon=False, fontsize=6)

            plt.title(f'Top {str(n)} {title}', fontsize=9, fontweight='bold')
            plt.xticks(rotation = 45, fontsize=6)
            plt.yticks(fontsize=7)
            #ax.set_xlim([mdates.epoch2num(df.index[0]), mdates.epoch2num(df.index[-1])])
            ax.xaxis.set_major_formatter(mdates.DateFormatter('%b %d'))
            ax.grid(axis='y')
            imagefile = f'{tenant["uniqueId"]}_{metric}.png'
            plt.savefig(imagefile, bbox_inches='tight')
            plt.clf()
            plt.close(fig)
            pdf.image(imagefile, x=pdf.l_margin, w=pdf.epw)
            os.remove(imagefile)

    pdf.ln(8)

    pdf.set_text_color(0,0,0)
    pdf.set_font('Helvetica', 'B', 11)
    pdf.cell(w=pdf.epw, txt=f'{title} Statistics', align='L', ln=1)
    pdf.ln(5)

    if (show_details):
        details = {}
        aggs = ['avg', 'min', 'max']
        for (agg_idx,agg) in enumerate(aggs):
            search = f'sort_desc({agg}_over_time({metric}[{range}:{span}]{offsetstring}))'
            data = metrics.promql_instance_search(openv, tenant['uniqueId'], search)
            for row in data:
                device_name = row['metric']['name']
                instance_name = row['metric']['instance']
                colname = device_name + " (" + instance_name + ")"
                if agg_idx == 0:
                    details[colname] = {}
                details[colname][agg] = float(row['value'][1])

 
        pdf.set_fill_color(224, 102, 54)
        pdf.set_text_color(255,255,255)
        pdf.set_font('Helvetica', 'B', 9)
        line_height = pdf.font_size * 2
        col_width = pdf.epw / 4 # distribute content evenly as baseline

        pdf.multi_cell(col_width*2.5, line_height, 'Device', border='LTB', ln=3, max_line_height=pdf.font_size, fill=True, align='L')
        pdf.multi_cell(col_width*.5, line_height, 'Minimum', border='TB', ln=3, max_line_height=pdf.font_size, fill=True, align='C')
        pdf.multi_cell(col_width*.5, line_height, 'Maximum', border='TB', ln=3, max_line_height=pdf.font_size, fill=True, align='C')
        pdf.multi_cell(col_width*.5, line_height, 'Average', border='TBR', ln=1, max_line_height=pdf.font_size, fill=True, align='C')

        pdf.set_text_color(0,0,0)
        pdf.set_font('Helvetica', '', 7)
        line_height = pdf.font_size * 2
        for idx, (col_name, aggs) in enumerate(details.items()):
            if idx%2 == 0:
                pdf.set_fill_color(252, 245, 233)
            else:
                pdf.set_fill_color(249, 234, 213)
            pdf.multi_cell(col_width*2.5, line_height, col_name, border='LTB', ln=3, max_line_height=pdf.font_size, fill=True, align='L')
            pdf.multi_cell(col_width*.5, line_height, "{:.2f}".format(aggs['min']), border='TB', ln=3, max_line_height=pdf.font_size, fill=True, align='C')
            pdf.multi_cell(col_width*.5, line_height, "{:.2f}".format(aggs['max']), border='TB', ln=3, max_line_height=pdf.font_size, fill=True, align='C')
            pdf.multi_cell(col_width*.5, line_height, "{:.2f}".format(aggs['avg']), border='TBR', ln=1, max_line_height=pdf.font_size, fill=True, align='C')          


def compute_dates(args):
    start = args.start
    end = args.end

    if start is None:
        start = datetime.today().astimezone()
        if start.month == 1 :
            start = start.replace(year=start.year-1, month=12, day=1,hour=0,minute=0,second=0,microsecond=0)
        else: 
            start = start.replace(month=start.month-1, day=1,hour=0,minute=0,second=0,microsecond=0)
    else:
        start = datetime.strptime(start,"%Y-%m-%d").astimezone()

    if end is None:
        if start.month == 12 :
            end = start.replace(year=start.year+1, month=1)
        else: 
            end = start.replace(month=start.month+1)
    else:
        end = datetime.strptime(end,"%Y-%m-%d").astimezone()

    ts = time.time()
    utc_offset = int((datetime.fromtimestamp(ts) - datetime.utcfromtimestamp(ts)).total_seconds())


    if hasattr(args, 'baseline_months_back'):
        bl_start = None
        bl_end = None
        baseline_months_back = args.baseline_months_back
        if baseline_months_back is None:
            baseline_months_back = 1
        bl_start = start + relativedelta(months=-baseline_months_back)
        bl_end = end + relativedelta(months=-baseline_months_back)
        return (start, end, bl_start, bl_end, utc_offset)
    else:
        return (start, end, None, None, utc_offset)

def pdf_table_from_df(pdf, df, includefields):

    if len(df) > 0:
        pdf.set_fill_color(224, 102, 54)
        pdf.set_text_color(255,255,255)
        pdf.set_font('Helvetica', 'B', 9)
        line_height = pdf.font_size * 4
        col_width = pdf.epw / len(includefields) # distribute content evenly as baseline
        for colidx,column in enumerate(includefields):
            border = 'TB'
            ln = 3
            if colidx == 0:
                border = 'LTB'
            elif colidx == len(includefields)-1:
                border = 'TBR'
                ln = 1
            pdf.multi_cell(col_width, line_height, column, border=border, ln=ln, max_line_height=pdf.font_size, fill=True, align='C')

        pdf.set_text_color(0,0,0)
        pdf.set_font('Helvetica', '', 7)
        line_height = pdf.font_size * 2
        iter = df.iterrows()
        try:
            while True:
                nextrec = next(iter)
                idx = nextrec[0]
                rec = nextrec[1]
                if idx%2 == 0:
                    pdf.set_fill_color(252, 245, 233)
                else:
                    pdf.set_fill_color(249, 234, 213)

                numrowsmax = int(math.ceil(max([pdf.get_string_width(str(rec[column]).replace('\u2013','-'))/col_width for column in includefields if not pd.isna(rec[column])])))
                this_line_height = line_height*numrowsmax
                for colidx, column in enumerate(includefields):
                    border = 'TB'
                    ln = 3
                    if colidx == 0:
                        border = 'LTB'
                    elif colidx == len(includefields)-1:
                        border = 'TBR'
                        ln = 1
                    colval = rec[column]
                    if pd.isna(colval):
                        colval = ""
                    else:
                        colval = str(colval)
                    if column == 'osName':
                        colval = re.sub(r' [.\d]+ Build.*','', colval)
                    pdf.multi_cell(col_width, this_line_height, colval.replace('\u2013','-'), border=border, ln=ln, max_line_height=pdf.font_size, fill=True, align='C')
        except StopIteration:
            pass   





import yaml
import requests
import sys

class OpsRampEnv:

    def __init__(self, envname, url, key, secret, partner_uniqueId, client_uniqueId, envfile='environments.yml', isSecure=True):
        if envname is not None:
            self.get_env(envname, envfile)
        else:
            self.getenv_nofile(url, key, secret, client_uniqueId, partner_uniqueId=None)

        self.isSecure = True
        if isinstance(isSecure, str) and (isSecure.lower() == 'false' or isSecure.lower() == 'no' or isSecure == '0'):
            self.isSecure = False
    
    def get_env(self, envname="", envfile=""):
        if hasattr(self, "env"):
            return self.env
        envstream = open(envfile, 'r')
        envs = yaml.safe_load(envstream)
        #print("Looking for environment named \"%s\" in %s." % (envname, envfile))
        filtered_envs = filter(lambda x: (x["name"] == envname), envs)
        try:
            self.env = next(filtered_envs)
            return self.env
        except StopIteration as err:
            raise Exception(f'No environment named "{envname}" found in {envfile}.')

    def getenv_nofile(self, url, key, secret, client_uniqueId, partner_uniqueId):
        if partner_uniqueId is None:
            partner_uniqueId = client_uniqueId
        if hasattr(self, "env"):
            return self.env
        else:
            self.env = {} 
            self.env['url'] = url
            self.env['partner'] = partner_uniqueId
            self.env['tenant'] = client_uniqueId
            self.env['client_id'] = key
            self.env['client_secret'] = secret
        return self.env


    def get_token(self):
        url = self.env['url'] + "/auth/oauth/token"
        payload = {
            'grant_type': 'client_credentials',
            'client_id': self.env['client_id'],
            'client_secret': self.env['client_secret']
        }
        headers = {
            'Content-Type': 'application/x-www-form-urlencoded',
            'Accept': 'application/json'
        }
        response = requests.request("POST", url, headers=headers, data=payload, verify=self.isSecure)
        #print(response.text)
        return response.json()['access_token']

    def get_tenants(self, queryString='activeStatus=true'):
        url = f'/api/v2/tenants/{self.env["partner"]}/clients/search'
        return self.do_get(url, queryString=queryString)


    def do_get(self, urlpath, page=1, queryString=None, searchQuery=None, countonly=False, attrId=None, paginate=True, pagebug=False):
        headers = {
            'Content-Type'      : 'application/json',
            'Accept'            : 'application/json',
            'Authorization'     : 'Bearer ' + self.get_token()
        }

        params = {}

        if paginate:
            params['pageSize'] = 100
            params['pageNo'] = page

        if countonly:
            params['pageSize'] = 1

        if queryString:
            params['queryString'] = queryString

        if searchQuery:
            params['searchQuery'] = searchQuery
            params['type'] = "resources"

        url = f'{self.env["url"]}{urlpath}'

        response = requests.request("GET", url, headers=headers, verify=self.isSecure, params=params)
        try:
            responseobj = response.json()
        except Exception as e:
            print(repr(response))
            sys.exit(1)

        if countonly:
            return int(responseobj['totalResults'])

        if "results" in responseobj:
            results = responseobj['results']
        else:
            results = responseobj

        if "nextPage" in responseobj and responseobj['nextPage'] and not pagebug:
            #print("Got %i %s from %s, proceeding to page %i" % (len(results), obtype, openv.env['name'], responseobj['nextPageNo']))
            return results + self.do_get(urlpath, page=responseobj['nextPageNo'],queryString=queryString, searchQuery=searchQuery, attrId=attrId)
        else:
            return results   